// This code example is created for educational purpose
// by Thorsten Thormaehlen (contact: www.thormae.de).
// It is distributed without any warranty.

import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;

import com.jogamp.opengl.GL3;
import com.jogamp.opengl.GLAutoDrawable;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLEventListener;
import com.jogamp.opengl.GLProfile;
import com.jogamp.opengl.awt.GLCanvas;
import javax.swing.JFrame;

import com.jogamp.common.nio.Buffers;
import com.jogamp.opengl.util.FPSAnimator;

class Renderer {

  public class LightSource {
    public float[] position = new float[3];
    public float[] color = new float[3];
    public float constantAttenuation, linearAttenuation, quadraticAttenuation;
    public float spotCutoff, spotExponent;
    public float[] spotDirection = new float[3];
    public float[] lookAt = new float[16];
    public float[] proj = new float[16];
    public float[] projLookAt = new float[16];
  };

  private class LightSourceUniformLoc {
    public int position;
    public int color;
    public int constantAttenuation, linearAttenuation, quadraticAttenuation;
    public int spotCutoff, spotExponent;
    public int spotDirection;
  };

  private class ShaderID {
    public int progID = 0;
    public int vertID = 0;
    public int fragID = 0;
  }

  public float t = 0.0f;
  public int modeVal = 1;
  public LightSource[] lights = new LightSource[3];
  public int activeLight;
  public boolean loadNext = false;

  private enum VAOs {
    Scene, Debug, numVAOs
  };

  private enum VBOs {
    SceneAll, DebugAll, numVBOs
  };

  private int[] vaoID = new int[VAOs.numVAOs.ordinal()];
  private int[] bufID = new int[VBOs.numVBOs.ordinal()];
  private int sceneVertNo = 0;
  private int vertexLoc = 0;
  private int texCoordLoc = 0;
  private int normalLoc = 0;
  private int projectionLoc = 0;
  private int modelviewLoc = 0;
  private int normalMatrixLoc = 0;
  private int modeLoc = 0;
  private float[] projection = new float[16];
  private float[] modelview = new float[16];
  private String filename = "./teapot.vbo";
  private int file = 0;
  private LightSourceUniformLoc[] lightsLoc = new LightSourceUniformLoc[3];
  private ShaderID[] shaders = new ShaderID[2];

  // for debug visualizations
  private int vertexDebugLoc;
  private int texCoordDebugLoc;
  private int colorDebugLoc;
  private int projectionDebugLoc;
  private int modelviewDebugLoc;
  private int modeDebugLoc;

  public void init(GLAutoDrawable d) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    gl.glEnable(GL3.GL_DEPTH_TEST);

    setupShaders(d);

    // create a Vertex Array Object (VAO)
    gl.glGenVertexArrays(VAOs.numVAOs.ordinal(), vaoID, 0);

    // generate a Vertex Buffer Object (VBO)
    gl.glGenBuffers(VBOs.numVBOs.ordinal(), bufID, 0);

    // binding the Triangle VAO
    gl.glBindVertexArray(vaoID[VAOs.Scene.ordinal()]);

    int perVertexFloats = (3 + 2 + 3);
    float data[] = loadVertexData(filename, perVertexFloats);

    sceneVertNo = data.length / (3 + 2 + 3);

    FloatBuffer sceneVertexFB = Buffers.newDirectFloatBuffer(data.length);
    sceneVertexFB.put(data);
    sceneVertexFB.flip();

    gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufID[VBOs.SceneAll.ordinal()]);
    gl.glBufferData(GL3.GL_ARRAY_BUFFER, sceneVertexFB.capacity() * Buffers.SIZEOF_FLOAT, sceneVertexFB,
        GL3.GL_STATIC_DRAW);

    int stride = (3 + 2 + 3) * Buffers.SIZEOF_FLOAT;
    int offset = 0;

    // position
    if (vertexLoc != -1) {
      gl.glVertexAttribPointer(vertexLoc, 3, GL3.GL_FLOAT, false, stride, offset);
      gl.glEnableVertexAttribArray(vertexLoc);
    }

    // texCoord
    if (texCoordLoc != -1) {
      offset = 0 + 3 * Buffers.SIZEOF_FLOAT;
      gl.glVertexAttribPointer(texCoordLoc, 2, GL3.GL_FLOAT, false, stride, offset);
      gl.glEnableVertexAttribArray(texCoordLoc);
    }

    // normal
    if (normalLoc != -1) {
      offset = 0 + (3 + 2) * Buffers.SIZEOF_FLOAT;
      gl.glVertexAttribPointer(normalLoc, 3, GL3.GL_FLOAT, false, stride, offset);
      gl.glEnableVertexAttribArray(normalLoc);
    }

    // the following is only to display debug information

    // binding the debug VAO
    gl.glBindVertexArray(vaoID[VAOs.Debug.ordinal()]);

    // In order to display the light's debug information (position and
    // orientiation),
    // the idea to render a cube [-1.0,1.0][-1.0,1.0][-1.0,1.0].
    // This cube is multiplied for drawing with the inverse "projection*modelview"
    // matrix
    // resulting in a "pyramid" (this is the inverse process of transforming
    // camera to clipping coordinates).

    float debugData[] = { -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 0.0f, 0.0f, 1.0f, // bottom
        -1.0f, -1.0f, 1.0f, -1.0f, -1.0f, 0.0f, 0.0f, 1.0f, 1.0f, -1.0f, 1.0f, -1.0f, -1.0f, 0.0f, 0.0f, 1.0f, 1.0f,
        -1.0f, -1.0f, -1.0f, -1.0f, 0.0f, 0.0f, 1.0f, -1.0f, 1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 0.0f, // top
        1.0f, 1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 0.0f, -1.0f,
        1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 0.0f, 1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, // right
        1.0f, -1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f,
        -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, // left
        -1.0f, 1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, -1.0f, 1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, -1.0f,
        -1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 1.0f, -1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 1.0f, 1.0f, // near
        1.0f, -1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 1.0f, 1.0f, -1.0f,
        1.0f, -1.0f, -1.0f, -1.0f, 1.0f, 1.0f, 1.0f, -1.0f, -1.0f, 1.0f, 0.0f, 0.0f, 1.0f, 1.0f, 1.0f, // far
        -1.0f, 1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, 1.0f, -1.0f,
        1.0f, 1.0f, 0.0f, 1.0f, 1.0f, 1.0f };

    FloatBuffer debugVertexFB = Buffers.newDirectFloatBuffer(debugData.length);
    debugVertexFB.put(debugData);
    debugVertexFB.flip();

    gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, bufID[VBOs.DebugAll.ordinal()]);
    gl.glBufferData(GL3.GL_ARRAY_BUFFER, debugVertexFB.capacity() * Buffers.SIZEOF_FLOAT, debugVertexFB,
        GL3.GL_STATIC_DRAW);

    // position
    offset = 0;
    gl.glVertexAttribPointer(vertexDebugLoc, 3, GL3.GL_FLOAT, false, stride, offset);
    gl.glEnableVertexAttribArray(vertexDebugLoc);

    // texCoord
    offset = 0 + 3 * Buffers.SIZEOF_FLOAT;
    gl.glVertexAttribPointer(texCoordDebugLoc, 2, GL3.GL_FLOAT, false, stride, offset);
    gl.glEnableVertexAttribArray(texCoordDebugLoc);

    // color
    offset = 0 + (3 + 2) * Buffers.SIZEOF_FLOAT;
    gl.glVertexAttribPointer(colorDebugLoc, 3, GL3.GL_FLOAT, false, stride, offset);
    gl.glEnableVertexAttribArray(colorDebugLoc);

    // create indices for the six sides of the cube
    // each side consists of 2 triangles with indices 0,1,2 and 0,2,3
    IntBuffer indicesData = Buffers.newDirectIntBuffer(6 * 6);
    for (int i = 0; i < 6; i++) {
      indicesData.put(0 + 4 * i);
      indicesData.put(1 + 4 * i);
      indicesData.put(2 + 4 * i);
      indicesData.put(0 + 4 * i);
      indicesData.put(2 + 4 * i);
      indicesData.put(3 + 4 * i);
    }
    indicesData.flip();

    // generating index VBO
    int[] indicesBufID = { 0 };
    gl.glGenBuffers(1, indicesBufID, 0);
    gl.glBindBuffer(GL3.GL_ELEMENT_ARRAY_BUFFER, indicesBufID[0]);
    gl.glBufferData(GL3.GL_ELEMENT_ARRAY_BUFFER, indicesData.capacity() * Buffers.SIZEOF_INT, indicesData,
        GL3.GL_STATIC_DRAW);

    gl.glBindVertexArray(0);
  }

  public void resize(GLAutoDrawable d, int w, int h) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    gl.glViewport(0, 0, w, h);

    // this function replaces gluPerspective
    mat4Perspective(projection, 45.0f, (float) w / (float) h, 0.5f, 40.0f);
    // mat4Print(projection);
  }

  public void display(GLAutoDrawable d) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL >= 3 graphics context

    gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    gl.glClear(GL3.GL_COLOR_BUFFER_BIT | GL3.GL_DEPTH_BUFFER_BIT);

    // camera orbits in the z=1.5 plane
    // and looks at the origin
    // mat4LookAt replaces gluLookAt
    double rad = Math.PI / 180.0f * t;

    mat4LookAt(modelview, 2.0f * (float) Math.cos(rad), 2.0f * (float) Math.sin(rad), 2.0f, // eye
        0.0f, 0.0f, 0.0f, // look at
        0.0f, 0.0f, 1.0f); // up

    float modelviewInv[] = new float[16];
    float normalmatrix[] = new float[16];
    mat4Invert(modelview, modelviewInv);
    mat4Transpose(modelviewInv, normalmatrix);
    updateLights();

    gl.glUseProgram(shaders[0].progID);

    // load the current projection and modelview matrix into the
    // corresponding UNIFORM variables of the shader
    gl.glUniformMatrix4fv(projectionLoc, 1, false, projection, 0);
    gl.glUniformMatrix4fv(modelviewLoc, 1, false, modelview, 0);
    gl.glUniformMatrix4fv(normalMatrixLoc, 1, false, normalmatrix, 0);
    gl.glUniform1i(modeLoc, modeVal);

    for (int i = 0; i < lightsLoc.length; i++) {
      gl.glUniform3fv(lightsLoc[i].position, 1, lights[i].position, 0);
      gl.glUniform3fv(lightsLoc[i].color, 1, lights[i].color, 0);
      gl.glUniform1f(lightsLoc[i].constantAttenuation, lights[i].constantAttenuation);
      gl.glUniform1f(lightsLoc[i].linearAttenuation, lights[i].linearAttenuation);
      gl.glUniform1f(lightsLoc[i].quadraticAttenuation, lights[i].quadraticAttenuation);
      gl.glUniform1f(lightsLoc[i].spotCutoff, lights[i].spotCutoff);
      gl.glUniform1f(lightsLoc[i].spotExponent, lights[i].spotExponent);
      gl.glUniform3fv(lightsLoc[i].spotDirection, 1, lights[i].spotDirection, 0);
    }

    // bind scene VAO
    gl.glBindVertexArray(vaoID[VAOs.Scene.ordinal()]);

    // render data
    gl.glDrawArrays(GL3.GL_TRIANGLES, 0, sceneVertNo);

    drawLightsDebug(d);

    gl.glFlush();
  }

  public void deleteAll(GLAutoDrawable d) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    gl.glDeleteVertexArrays(VAOs.numVAOs.ordinal(), vaoID, 0);
    gl.glDeleteBuffers(VBOs.numVBOs.ordinal(), bufID, 0);
    for (int i = 0; i < shaders.length; i++) {
      gl.glDeleteProgram(shaders[i].progID);
      gl.glDeleteShader(shaders[i].vertID);
      gl.glDeleteShader(shaders[i].fragID);
    }
  }

  private void updateLights() {
    for (int i = 0; i < 3; i++) {
      float nullVec[] = { 0.0f, 0.0f, 0.0f };
      float nullTrans[] = new float[3];
      float res[] = new float[3];
      mat4MultVec3(modelview, nullVec, nullTrans);
      res[0] = nullTrans[0] - lights[i].position[0];
      res[1] = nullTrans[1] - lights[i].position[1];
      res[2] = nullTrans[2] - lights[i].position[2];
      vec3Normalize(res);
      lights[i].spotDirection[0] = res[0];
      lights[i].spotDirection[1] = res[1];
      lights[i].spotDirection[2] = res[2];
      mat4LookAt(lights[i].lookAt, lights[i].position[0], lights[i].position[1], lights[i].position[2], // eye
          nullTrans[0], nullTrans[1], nullTrans[2], // look at
          0.0f, 1.0f, 0.0f); // up
      mat4Perspective(lights[i].proj, 30.0f, 9.0f / 9.0f, 0.001f, 0.3f);
      mat4Multiply(lights[i].proj, lights[i].lookAt, lights[i].projLookAt);
    }
  }

  void createLights() {

    for (int i = 0; i < lights.length; i++) {
      lights[i] = new LightSource();
      lights[i].position[0] = 0.0f;
      lights[i].position[1] = 0.0f;
      lights[i].position[2] = -1.0f;
      lights[i].color[0] = 1.0f;
      lights[i].color[1] = 1.0f;
      lights[i].color[2] = 1.0f;
      lights[i].constantAttenuation = 1.0f;
      lights[i].linearAttenuation = 0.0f;
      lights[i].quadraticAttenuation = 0.0f;
      lights[i].spotCutoff = 0.0f;
      lights[i].spotExponent = 60.0f;
    }

    lights[0].position[0] = 0.8f;
    lights[0].position[1] = -0.2f;
    lights[0].position[2] = -2.1f;

    lights[1].position[0] = -1.1f;
    lights[1].position[1] = -0.4f;
    lights[1].position[2] = -3.0f;

    lights[2].position[0] = -0.3f;
    lights[2].position[1] = -0.6f;
    lights[2].position[2] = -2.2f;
  }

  private void drawLightsDebug(GLAutoDrawable d) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context

    // In order to display the light's debug information (position and
    // orientiation),
    // the idea to render a cube [-1.0,1.0][-1.0,1.0][-1.0,1.0].
    // This cube is multiplied for drawing with the inverse "projection*modelview"
    // matrix
    // resulting in a "pyramid" (this is the inverse process of transforming
    // camera to clipping coordinates).

    // bind debug VAO
    gl.glBindVertexArray(vaoID[VAOs.Debug.ordinal()]);

    gl.glUseProgram(shaders[1].progID);
    gl.glUniformMatrix4fv(projectionDebugLoc, 1, false, projection, 0);

    for (int i = 0; i < lights.length; i++) {
      float lightInv[] = new float[16];
      mat4Invert(lights[i].projLookAt, lightInv);
      gl.glUniformMatrix4fv(modelviewDebugLoc, 1, false, lightInv, 0);
      if (modeDebugLoc != -1)
        gl.glUniform1i(modeDebugLoc, modeVal);

      // render data
      gl.glDrawElements(GL3.GL_TRIANGLES, 6 * 6, GL3.GL_UNSIGNED_INT, 0);
    }

    gl.glUseProgram(0);
  }

  private void setupShaders(GLAutoDrawable ad) {
    GL3 gl = ad.getGL().getGL3(); // get the OpenGL 3 graphics context

    // blinn phong shader
    ShaderID s = createShader(ad, "./BlinnPhong.vert", "./BlinnPhong.frag");
    // retrieve the location of the IN variables of the vertex shader
    vertexLoc = gl.glGetAttribLocation(s.progID, "inputPosition");
    texCoordLoc = gl.glGetAttribLocation(s.progID, "inputTexCoord");
    normalLoc = gl.glGetAttribLocation(s.progID, "inputNormal");

    // retrieve the location of the UNIFORM variables of the vertex shader
    projectionLoc = gl.glGetUniformLocation(s.progID, "projection");
    modelviewLoc = gl.glGetUniformLocation(s.progID, "modelview");
    normalMatrixLoc = gl.glGetUniformLocation(s.progID, "normalMat");
    modeLoc = gl.glGetUniformLocation(s.progID, "mode");

    for (int i = 0; i < lightsLoc.length; i++) {
      lightsLoc[i] = new LightSourceUniformLoc();
      String b = String.valueOf(i);
      String a = "lights[";
      lightsLoc[i].position = gl.glGetUniformLocation(s.progID, a + b + "].position");
      lightsLoc[i].color = gl.glGetUniformLocation(s.progID, a + b + "].color");
      lightsLoc[i].constantAttenuation = gl.glGetUniformLocation(s.progID, a + b + "].constantAttenuation");
      lightsLoc[i].linearAttenuation = gl.glGetUniformLocation(s.progID, a + b + "].linearAttenuation");
      lightsLoc[i].quadraticAttenuation = gl.glGetUniformLocation(s.progID, a + b + "].quadraticAttenuation");
      lightsLoc[i].spotCutoff = gl.glGetUniformLocation(s.progID, a + b + "].spotCutoff");
      lightsLoc[i].spotExponent = gl.glGetUniformLocation(s.progID, a + b + "].spotExponent");
      lightsLoc[i].spotDirection = gl.glGetUniformLocation(s.progID, a + b + "].spotDirection");
    }

    shaders[0] = s;

    // debug information shader
    ShaderID d = createShader(ad, "./debug.vert", "./debug.frag");

    // retrieve the location of the IN variables of the vertex shaders
    vertexDebugLoc = gl.glGetAttribLocation(d.progID, "inputPosition");
    texCoordDebugLoc = gl.glGetAttribLocation(d.progID, "inputTexCoord");
    colorDebugLoc = gl.glGetAttribLocation(d.progID, "inputColor");

    // retrieve the location of the UNIFORM variables of the shaders
    projectionDebugLoc = gl.glGetUniformLocation(d.progID, "projection");
    modelviewDebugLoc = gl.glGetUniformLocation(d.progID, "modelview");
    modeDebugLoc = gl.glGetUniformLocation(d.progID, "mode");

    shaders[1] = d;
  }

  private ShaderID createShader(GLAutoDrawable d, String vertSrc, String fragSrc) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    ShaderID s = new ShaderID();

    s.vertID = gl.glCreateShader(GL3.GL_VERTEX_SHADER);
    s.fragID = gl.glCreateShader(GL3.GL_FRAGMENT_SHADER);

    String[] vs = loadShaderSrc(vertSrc);
    String[] fs = loadShaderSrc(fragSrc);

    gl.glShaderSource(s.vertID, 1, vs, null, 0);
    gl.glShaderSource(s.fragID, 1, fs, null, 0);

    // compile the shader
    gl.glCompileShader(s.vertID);
    gl.glCompileShader(s.fragID);

    // check for errors
    printShaderInfoLog(d, s.vertID);
    printShaderInfoLog(d, s.fragID);

    // create program and attach shaders
    s.progID = gl.glCreateProgram();
    gl.glAttachShader(s.progID, s.vertID);
    gl.glAttachShader(s.progID, s.fragID);

    // "outColor" is a user-provided OUT variable
    // of the fragment shader.
    // Its output is bound to the first color buffer
    // in the framebuffer
    gl.glBindFragDataLocation(s.progID, 0, "outputColor");

    // link the program
    gl.glLinkProgram(s.progID);
    // output error messages
    printProgramInfoLog(d, s.progID);

    return s;
  }

  private String[] loadShaderSrc(String name) {
    StringBuilder sb = new StringBuilder();
    try {
      InputStream is = getClass().getResourceAsStream(name);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line;
      while ((line = br.readLine()) != null) {
        sb.append(line);
        sb.append('\n');
      }
      is.close();
    } catch (Exception e) {
      e.printStackTrace();
    }
    return new String[] { sb.toString() };
  }

  private void printShaderInfoLog(GLAutoDrawable d, int obj) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    int infoLogLength;
    gl.glGetShaderiv(obj, GL3.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      ByteBuffer byteBuffer = ByteBuffer.allocate(infoLogLength);
      gl.glGetShaderInfoLog(obj, infoLogLength, infoLogLengthBuf, byteBuffer);
      for (byte b : byteBuffer.array()) {
        System.err.print((char) b);
      }
    }
  }

  private void printProgramInfoLog(GLAutoDrawable d, int obj) {
    GL3 gl = d.getGL().getGL3(); // get the OpenGL 3 graphics context
    IntBuffer infoLogLengthBuf = IntBuffer.allocate(1);
    int infoLogLength;
    gl.glGetProgramiv(obj, GL3.GL_INFO_LOG_LENGTH, infoLogLengthBuf);
    infoLogLength = infoLogLengthBuf.get(0);
    if (infoLogLength > 0) {
      ByteBuffer byteBuffer = ByteBuffer.allocate(infoLogLength);
      gl.glGetProgramInfoLog(obj, infoLogLength, infoLogLengthBuf, byteBuffer);
      for (byte b : byteBuffer.array()) {
        System.err.print((char) b);
      }
    }
  }

  private float[] loadVertexData(String filename, int perVertexFloats) {

    float[] floatArray = new float[0];
    // read vertex data from file
    int vertSize = 0;
    try {
      InputStream is = getClass().getResourceAsStream(filename);
      BufferedReader br = new BufferedReader(new InputStreamReader(is));
      String line = br.readLine();
      if (line != null) {
        vertSize = Integer.parseInt(line);
        floatArray = new float[vertSize];
      }
      int i = 0;
      while ((line = br.readLine()) != null && i < floatArray.length) {
        floatArray[i] = Float.parseFloat(line);
        i++;
      }
      if (i != vertSize || (vertSize % perVertexFloats) != 0) {
        floatArray = new float[0];
      }
      br.close();
    } catch (FileNotFoundException e) {
      System.out.println("Can not find vbo data file " + filename);
    } catch (IOException e) {
      e.printStackTrace();
    }
    return floatArray;
  }

  // the following functions are some matrix and vector helpers
  // they work for this demo but in general it is recommended
  // to use more advanced matrix libraries
  private float vec3Dot(float[] a, float[] b) {
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
  }

  private void vec3Add(float[] a, float[] b, float[] res) {
    res[0] = a[0] + b[0];
    res[1] = a[1] + b[1];
    res[2] = a[2] + b[2];
  }

  private void vec3Cross(float[] a, float[] b, float[] res) {
    res[0] = a[1] * b[2] - b[1] * a[2];
    res[1] = a[2] * b[0] - b[2] * a[0];
    res[2] = a[0] * b[1] - b[0] * a[1];
  }

  private void vec3Normalize(float[] a) {
    float mag = (float) Math.sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
    a[0] /= mag;
    a[1] /= mag;
    a[2] /= mag;
  }

  private void mat4Identity(float[] a) {
    for (int i = 0; i < 16; ++i)
      a[i] = 0.0f;
    for (int i = 0; i < 4; ++i)
      a[i + i * 4] = 1.0f;
  }

  private void mat4Scale(float[] a, float sx, float sy, float sz) {
    for (int i = 0; i < 16; ++i)
      a[i] = 0.0f;
    a[0 + 0 * 4] = sx;
    a[1 + 1 * 4] = sy;
    a[2 + 2 * 4] = sz;
    a[3 + 3 * 4] = 1.0f;
  }

  private void mat4Multiply(float[] a, float[] b, float[] res) {
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        res[j * 4 + i] = 0.0f;
        for (int k = 0; k < 4; ++k) {
          res[j * 4 + i] += a[k * 4 + i] * b[j * 4 + k];
        }
      }
    }
  }

  private void mat4MultVec3(float[] mat, float[] vec, float[] res) {
    for (int i = 0; i < 3; ++i) {
      res[i] = 0.0f;
      for (int k = 0; k < 3; ++k) {
        res[i] += mat[i + k * 4] * vec[k];
      }
      res[i] += mat[i + 3 * 4] * 1.0f;
    }
  }

  private void mat4Perspective(float[] a, float fov, float aspect, float zNear, float zFar) {
    float f = 1.0f / (float) (Math.tan(fov / 2.0f * (Math.PI / 180.0f)));
    mat4Identity(a);
    a[0] = f / aspect;
    a[1 * 4 + 1] = f;
    a[2 * 4 + 2] = (zFar + zNear) / (zNear - zFar);
    a[3 * 4 + 2] = (2.0f * zFar * zNear) / (zNear - zFar);
    a[2 * 4 + 3] = -1.0f;
    a[3 * 4 + 3] = 0.0f;
  }

  private void mat4LookAt(float[] viewMatrix, float eyeX, float eyeY, float eyeZ, float centerX, float centerY,
      float centerZ, float upX, float upY, float upZ) {

    float dir[] = new float[3];
    float right[] = new float[3];
    float up[] = new float[3];
    float eye[] = new float[3];

    up[0] = upX;
    up[1] = upY;
    up[2] = upZ;
    eye[0] = eyeX;
    eye[1] = eyeY;
    eye[2] = eyeZ;

    dir[0] = centerX - eyeX;
    dir[1] = centerY - eyeY;
    dir[2] = centerZ - eyeZ;
    vec3Normalize(dir);
    vec3Cross(dir, up, right);
    vec3Normalize(right);
    vec3Cross(right, dir, up);
    vec3Normalize(up);
    // first row
    viewMatrix[0] = right[0];
    viewMatrix[4] = right[1];
    viewMatrix[8] = right[2];
    viewMatrix[12] = -vec3Dot(right, eye);
    // second row
    viewMatrix[1] = up[0];
    viewMatrix[5] = up[1];
    viewMatrix[9] = up[2];
    viewMatrix[13] = -vec3Dot(up, eye);
    // third row
    viewMatrix[2] = -dir[0];
    viewMatrix[6] = -dir[1];
    viewMatrix[10] = -dir[2];
    viewMatrix[14] = vec3Dot(dir, eye);
    // forth row
    viewMatrix[3] = 0.0f;
    viewMatrix[7] = 0.0f;
    viewMatrix[11] = 0.0f;
    viewMatrix[15] = 1.0f;
  }

  private void mat4Print(float[] a) {
    // opengl uses column major order
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        System.out.print(a[j * 4 + i] + " ");
      }
      System.out.println(" ");
    }
  }

  private void mat4Transpose(float[] a, float[] transposed) {
    int t = 0;
    for (int i = 0; i < 4; ++i) {
      for (int j = 0; j < 4; ++j) {
        transposed[t++] = a[j * 4 + i];
      }
    }
  }

  private boolean mat4Invert(float[] m, float[] inverse) {
    float inv[] = new float[16];
    inv[0] = m[5] * m[10] * m[15] - m[5] * m[11] * m[14] - m[9] * m[6] * m[15] + m[9] * m[7] * m[14]
        + m[13] * m[6] * m[11] - m[13] * m[7] * m[10];
    inv[4] = -m[4] * m[10] * m[15] + m[4] * m[11] * m[14] + m[8] * m[6] * m[15] - m[8] * m[7] * m[14]
        - m[12] * m[6] * m[11] + m[12] * m[7] * m[10];
    inv[8] = m[4] * m[9] * m[15] - m[4] * m[11] * m[13] - m[8] * m[5] * m[15] + m[8] * m[7] * m[13]
        + m[12] * m[5] * m[11] - m[12] * m[7] * m[9];
    inv[12] = -m[4] * m[9] * m[14] + m[4] * m[10] * m[13] + m[8] * m[5] * m[14] - m[8] * m[6] * m[13]
        - m[12] * m[5] * m[10] + m[12] * m[6] * m[9];
    inv[1] = -m[1] * m[10] * m[15] + m[1] * m[11] * m[14] + m[9] * m[2] * m[15] - m[9] * m[3] * m[14]
        - m[13] * m[2] * m[11] + m[13] * m[3] * m[10];
    inv[5] = m[0] * m[10] * m[15] - m[0] * m[11] * m[14] - m[8] * m[2] * m[15] + m[8] * m[3] * m[14]
        + m[12] * m[2] * m[11] - m[12] * m[3] * m[10];
    inv[9] = -m[0] * m[9] * m[15] + m[0] * m[11] * m[13] + m[8] * m[1] * m[15] - m[8] * m[3] * m[13]
        - m[12] * m[1] * m[11] + m[12] * m[3] * m[9];
    inv[13] = m[0] * m[9] * m[14] - m[0] * m[10] * m[13] - m[8] * m[1] * m[14] + m[8] * m[2] * m[13]
        + m[12] * m[1] * m[10] - m[12] * m[2] * m[9];
    inv[2] = m[1] * m[6] * m[15] - m[1] * m[7] * m[14] - m[5] * m[2] * m[15] + m[5] * m[3] * m[14] + m[13] * m[2] * m[7]
        - m[13] * m[3] * m[6];
    inv[6] = -m[0] * m[6] * m[15] + m[0] * m[7] * m[14] + m[4] * m[2] * m[15] - m[4] * m[3] * m[14]
        - m[12] * m[2] * m[7] + m[12] * m[3] * m[6];
    inv[10] = m[0] * m[5] * m[15] - m[0] * m[7] * m[13] - m[4] * m[1] * m[15] + m[4] * m[3] * m[13]
        + m[12] * m[1] * m[7] - m[12] * m[3] * m[5];
    inv[14] = -m[0] * m[5] * m[14] + m[0] * m[6] * m[13] + m[4] * m[1] * m[14] - m[4] * m[2] * m[13]
        - m[12] * m[1] * m[6] + m[12] * m[2] * m[5];
    inv[3] = -m[1] * m[6] * m[11] + m[1] * m[7] * m[10] + m[5] * m[2] * m[11] - m[5] * m[3] * m[10] - m[9] * m[2] * m[7]
        + m[9] * m[3] * m[6];
    inv[7] = m[0] * m[6] * m[11] - m[0] * m[7] * m[10] - m[4] * m[2] * m[11] + m[4] * m[3] * m[10] + m[8] * m[2] * m[7]
        - m[8] * m[3] * m[6];
    inv[11] = -m[0] * m[5] * m[11] + m[0] * m[7] * m[9] + m[4] * m[1] * m[11] - m[4] * m[3] * m[9] - m[8] * m[1] * m[7]
        + m[8] * m[3] * m[5];
    inv[15] = m[0] * m[5] * m[10] - m[0] * m[6] * m[9] - m[4] * m[1] * m[10] + m[4] * m[2] * m[9] + m[8] * m[1] * m[6]
        - m[8] * m[2] * m[5];

    float det = m[0] * inv[0] + m[1] * inv[4] + m[2] * inv[8] + m[3] * inv[12];
    if (det == 0)
      return false;
    det = 1.0f / det;
    for (int i = 0; i < 16; i++)
      inverse[i] = inv[i] * det;
    return true;
  }

}

class MyGui extends JFrame implements GLEventListener {

  private Renderer renderer;

  public void createGUI() {
    setTitle("Shader Many Modifiable Lights");
    setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

    GLProfile glp = GLProfile.getDefault();
    GLCapabilities caps = new GLCapabilities(glp);
    GLCanvas canvas = new GLCanvas(caps);
    setSize(320, 320);

    getContentPane().add(canvas);
    final FPSAnimator ani = new FPSAnimator(canvas, 60, true);
    canvas.addGLEventListener(this);
    setVisible(true);
    renderer = new Renderer();
    renderer.createLights();

    canvas.addKeyListener(new KeyAdapter() {
      public void keyPressed(KeyEvent event) {
        boolean redraw = false;
        String modeStr = "";
        int l = renderer.activeLight;
        switch (event.getKeyCode()) {
        case '1':
          renderer.modeVal = 1;
          redraw = true;
          break;
        case '2':
          renderer.modeVal = 2;
          redraw = true;
          break;
        case '3':
          renderer.modeVal = 3;
          redraw = true;
          break;
        case '4':
          renderer.modeVal = 4;
          redraw = true;
          break;
        case '5':
          renderer.modeVal = 5;
          redraw = true;
          break;
        case KeyEvent.VK_PLUS:
          renderer.activeLight++;
          if (renderer.activeLight >= renderer.lights.length)
            renderer.activeLight = 0;
          redraw = true;
          break;
        case '-':
          renderer.activeLight--;
          if (renderer.activeLight < 0)
            renderer.activeLight = renderer.lights.length - 1;
          redraw = true;
          break;
        case 'd':
        case 'D':
        case KeyEvent.VK_RIGHT:
          renderer.lights[l].position[0] += 0.1f;
          redraw = true;
          break;
        case 'a':
        case 'A':
        case KeyEvent.VK_LEFT:
          renderer.lights[l].position[0] -= 0.1f;
          redraw = true;
          break;
        case 's':
        case 'S':
        case KeyEvent.VK_DOWN:
          renderer.lights[l].position[1] -= 0.1f;
          redraw = true;
          break;
        case 'w':
        case 'W':
        case KeyEvent.VK_UP:
          renderer.lights[l].position[1] += 0.1f;
          redraw = true;
          break;
        case ' ':
          renderer.lights[l].position[2] -= 0.1f;
          redraw = true;
          break;
        case 'c':
        case 'C':
          renderer.lights[l].position[2] += 0.1f;
          redraw = true;
          break;
        case 'm':
        case 'M':
          renderer.lights[l].linearAttenuation += 0.01f;
          redraw = true;
          break;
        case 'n':
        case 'N':
          renderer.lights[l].linearAttenuation -= 0.01f;
          if (renderer.lights[l].linearAttenuation <= 0.0f)
            renderer.lights[0].linearAttenuation = 0.01f;
          redraw = true;
          break;
        case 'k':
        case 'K':
          renderer.lights[l].spotCutoff += 0.001f;
          if (renderer.lights[l].spotCutoff > 1.0f)
            renderer.lights[l].spotCutoff = 1.0f;
          redraw = true;
          break;
        case 'j':
        case 'J':
          renderer.lights[l].spotCutoff -= 0.001f;
          if (renderer.lights[l].spotCutoff < 0.0f)
            renderer.lights[l].spotCutoff = 0.0f;
          redraw = true;
          break;
        case 'o':
        case 'O':
          renderer.lights[l].spotExponent += 1.0f;
          redraw = true;
          break;
        case 'i':
        case 'I':
          renderer.lights[l].spotExponent -= 1.0f;
          if (renderer.lights[l].spotExponent < 0.0)
            renderer.lights[0].spotExponent = 0.0f;
          redraw = true;
          break;
        }
        if (redraw) {
          l = renderer.activeLight;
          modeStr = "Light " + l + ": lightPos=" + renderer.lights[l].position[0] + "," + renderer.lights[l].position[1]
              + "," + renderer.lights[l].position[2] + " attenuation=" + renderer.lights[l].linearAttenuation
              + " cutoff=" + renderer.lights[l].spotCutoff + " exponent=" + renderer.lights[l].spotExponent;
          setTitle(modeStr);
        }
      }
    });

    ani.start();
  }

  @Override
  public void init(GLAutoDrawable d) {
    renderer.init(d);
  }

  @Override
  public void reshape(GLAutoDrawable d, int x, int y, int width, int height) {
    renderer.resize(d, width, height);
  }

  @Override
  public void display(GLAutoDrawable d) {
    float offset = 1.0f;
    renderer.t += offset;
    renderer.display(d);
  }

  @Override
  public void dispose(GLAutoDrawable d) {
    renderer.deleteAll(d);
  }
}

public class ShaderPhongSpot {
  public static void main(String[] args) {
    javax.swing.SwingUtilities.invokeLater(new Runnable() {
      public void run() {
        MyGui myGUI = new MyGui();
        myGUI.createGUI();
      }
    });
  }
}
